/////////////////////////////////////
//                                 //
//         PanEngine API           //
//                                 //
//         Version 2020            //
//                                 //
//        CompuTherm, LLC          //
//      Copyright, 2012-2020       //
//                                 //
//       www.computherm.com        //
//                                 //
/////////////////////////////////////

//==================================//
//                                  //
//    PanEngine Test Example 1      //
//                                  //
//        Point Equilibrium         //
//         (Global Stable)          //
//        (Al-Mg-Zn System)         //
//----------------------------------//
//  Define a PanEngine user pointer //
//  Set PanEngine configuration     //
//  Read a database                 //
//  Define a subsystem              //
//  Set calculation condition       //
//  Get physical properties         //
//  Get Hessian matrix, eigenvalues //
//      and eigenvectors            // 
//  Get driving force of phase      //
//  Delete a PanEngine user pointer //
//                                  //
//==================================//


#include "std.h"
#include "PanEngineX.h"


void set_calculation_condition_1(Pan_Calculation& calc, pair<string, string> db) {

	calc.m_name="Point Equilibrium";
	calc.m_calc_type = P_CALC_POINT;

	// databases
	shared_ptr<Pan_Database> d_b = make_shared<Pan_Database>();
	d_b->m_file_name = db.first;
	d_b->m_type = db.second;
	calc.m_databases.push_back(pair<string, shared_ptr<Pan_Database>> (d_b->m_file_name, d_b));

	// units
	shared_ptr<Pan_Unit> unit_T = make_shared<Pan_Unit>();
	unit_T->m_name = "T";
	unit_T->m_unit = P_T_K;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_T->m_name, unit_T));

	shared_ptr<Pan_Unit> unit_P = make_shared<Pan_Unit>();
	unit_P->m_name = "P";
	unit_P->m_unit = P_P_atm;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_P->m_name, unit_P));

	// set unit for component amount
	shared_ptr<Pan_Unit> unit_m = make_shared<Pan_Unit>();
	unit_m->m_name = "n";
	unit_m->m_unit = P_M_mole;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_m->m_name, unit_m));

	// system
	calc.m_system.push_back(make_shared<Pan_System>());
	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];
	
	// components
	shared_ptr<Pan_Component> comp1 = make_shared<Pan_Component>();
	comp1->m_name = "Al";
	comp1->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp1);

	shared_ptr<Pan_Component> comp2 = make_shared<Pan_Component>();
	comp2->m_name = "Mg";
	comp2->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp2);

	shared_ptr<Pan_Component> comp3 = make_shared<Pan_Component>();
	comp3->m_name = "Zn";
	comp3->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp3);

	// phases
	shared_ptr<Pan_Phase> phase0 = make_shared<Pan_Phase>();
	phase0->m_name = "*";
	phase0->m_status = P_PHASE_SUSPENDED;
	sys_ptr->m_phases.push_back(phase0);

	shared_ptr<Pan_Phase> phase1 = make_shared<Pan_Phase>();
	phase1->m_name = "Liquid";
	phase1->m_status = P_PHASE_ENTERED;
	sys_ptr->m_phases.push_back(phase1);

	shared_ptr<Pan_Phase> phase2 = make_shared<Pan_Phase>();
	phase2->m_name = "FCC_A1";
	phase2->m_status = P_PHASE_ENTERED;
	sys_ptr->m_phases.push_back(phase2);

	shared_ptr<Pan_Phase> phase3 = make_shared<Pan_Phase>();
	phase3->m_name = "BCC_A2";
	phase3->m_status = P_PHASE_DORMANT;
	sys_ptr->m_phases.push_back(phase3);

	// calculation condition	
	// calculate driving forces for all unsuspended phases
	sys_ptr->m_condition->m_driving_force = false;
	sys_ptr->m_condition->m_physical_property = true;

	sys_ptr->m_condition->m_solvent = string("Al");
	// Al is set to be solvent component
	// Mg and Zn will be the solutes
	// x(Mg) and x(Zn) will be the independet variables in Hessian

	// point
	shared_ptr<Pan_Point> pt1 = make_shared<Pan_Point>();
	shared_ptr<Pan_Statespace> st;
	st = make_shared<Pan_Statespace>();
	st->m_T = 600;
	st->m_P = 1;
	st->m_n.insert(pair<string, double> ("Al", 0.2));
	st->m_n.insert(pair<string, double> ("Mg", 0.3));
	st->m_n.insert(pair<string, double> ("Zn", 0.5));

	pt1->m_statespace = st;

	sys_ptr->m_points.push_back(pt1);
}

int PanEngineX_Test_1()
{
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
	cout << "|  Test Example 1 : Point Equilibrium  |\n";
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";


	//=================================
	// Define a PanEngineX user pointer
	//=================================
	PanEngine* user;
	char msg[256];
	string s;

	cout << "Defining user pointer:" << endl;
	user = definePanEngineUser(msg);

	if (user == NULL) 
	{
		cout << "Error in defineUser" << endl;
		return 0;
	}

	//============================
	// Set PanEngine Configuration
	//============================
	// set "case_sensitive" before importing database
	cout << "Set PanEngine configuration:" << endl;
	map<string, string> config;
	config["case_sensitive"] = "true";
	s = user->pe_set_configuration(config);

	//==========================================
	// Load Thermodynamic Database in tdb Format
	//==========================================
	cout << "Reading database file:" << endl;

	pair<string, string> db = make_pair<string, string>(string("AlMgZn.tdb"), string("tdb"));

	s = user->pe_import_database(db);

	if (!s.empty())
	{
		cout << "Error in open database [" << s << "]\n";
		return 0;
	}

	//==============================
	// Get Component and Phase Names
	//==============================
	vector<string> comp_name;
	vector<string> phase_name;
	s = user->pe_get_component_names(db, comp_name);
	s = user->pe_get_phase_names(db, phase_name);

	cout << "comp_name =\n";
	for_each(comp_name.begin(), comp_name.end(), [&](string c_name) {	cout << "\t" << c_name; }	);
	cout << endl;
	cout << "phase_name =\n";
	for_each(phase_name.begin(), phase_name.end(), [&](string p_name) {	cout << "\t" << p_name << endl; }	);
	cout << endl;

	//==============================
	// set up calculation conditions
	//==============================
	Pan_Calculation calc;
	set_calculation_condition_1(calc, db);
	shared_ptr<P_Point> calc_point = make_shared<P_Point>();

	s = user->pe_set_calculation_condition(calc);
	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];

	if (!s.empty()) {
		cout << "Error in calculation condition [" << s << "]\n";
		return 0;
	}

	vector<string> prop_v;
	prop_v.push_back("Vm");
	prop_v.push_back("density");
	prop_v.push_back("MW");

	//==================
	// Point Calculation
	//==================
	cout << "\n\n========= Point 1 =========\n";
	// T = 600K x[Al]=0.2 x[Mg]=0.3 x[Zn]=0.5
	s = user->pe_calc_point_global(calc_point);
	s = user->pe_update_properties(calc_point, false, prop_v);
	calc_point->print_debug(cout);

	for (auto ppt = calc_point->m_ppt.begin(); ppt != calc_point->m_ppt.end(); ++ppt) {
		cout << "Phase Name = " << (*ppt)->m_phase_name << endl;
		if ((*ppt)->m_physical.first == true) {
			cout << "  Physical Properties:\n";
			cout << "    Molecular weight = " << (*ppt)->m_physical.second.get_value("MW") << " g/mole" << endl;
			cout << "    Molar volume     = " << (*ppt)->m_physical.second.get_value("Vm") << " cm^3/mole" << endl;
			cout << "    Density          = " << (*ppt)->m_physical.second.get_value("density") << " g/cm^3" << endl << endl;
		}

		if ((*ppt)->m_Hess.first == true) {
			cout << "  Hessian Matrix of G:" << endl;
			cout << "    HSN =";
			auto HSN_MTX = (*ppt)->m_Hess.second.m_Hessian;
			for (auto i = HSN_MTX.begin(); i != HSN_MTX.end(); ++i) {
				if (i != HSN_MTX.begin()) cout << "         ";
				for (auto j = (*i).begin(); j != (*i).end(); ++j) {
					cout << " <" << get<0>((*j)) << ", " << get<1>((*j)) << ", " << get<2>((*j)) << ">  ";
				}
				cout << endl;
			}

			cout << "  Determinant of Hessian Matrix of G:" << endl;
			cout << "    |HSN| = " << (*ppt)->m_Hess.second.m_HSN << endl;

			cout << "  Eigenvalues and Eigenvectors of Hessian Matrix of G:" << endl;
			auto eigen = (*ppt)->m_Hess.second.m_eigen;
			for (auto it = eigen.begin(); it != eigen.end(); ++it) {
				cout << "    eVal = " << (*it).first << "  eVec:";
				for (auto v = (*it).second.begin(); v != (*it).second.end(); ++v) {
					cout << " <" << (*v).first << ", " << (*v).second << ">";
				}
				cout << endl;
			}
			cout << endl;
		}
	}

	cout << "\n\n========= Point 2 =========\n";
	// calculate driving forces for all unsuspended phases
	calc_point->m_st->m_T -= 200.0;
	
	sys_ptr->m_condition->m_driving_force = true;

	s = user->pe_calc_point_global_with_initial_point(calc_point);

	calc_point->print_debug(cout);

	// get driving force for a phase
	string ph_name = "BCC_A2";
	cout << "Driving Force for " << ph_name << " = " << calc_point->get_driving_force(ph_name) << endl;

	cout << "\n\n========= End of Example 1 =========\n\n";

	//===============================
	// Delete PanEngineX user pointer
	//===============================
	deletePanEngineUser(user);

	return 1;
}

