/////////////////////////////////////
//                                 //
//         PanEngine API           //
//                                 //
//         Version 2020            //
//                                 //
//        CompuTherm, LLC          //
//      Copyright, 2012-2020       //
//                                 //
//       www.computherm.com        //
//                                 //
/////////////////////////////////////

//==================================//
//                                  //
//    PanEngine Test Example 2      //
//                                  //
//       Point Equilibria           //
//      for Multiple Points         //
//        (Global Stable)           //
//       (Al-Mg-Zn System)          //
//                                  //
//==================================//


#include "std.h"
#include "PanEngineX.h"


bool valid_x(double& x) {
	double eps=1e-10;
	if (fabs(x) < eps) {
		x = 0;
	} else
	if (fabs(1-x) < eps) {		
		x = 1;
	}
	if (x < 0 || x > 1) return false;
	return true;
}

void set_calculation_condition_2(Pan_Calculation& calc, pair<string, string> db) {

	calc.m_name="Equilibria for Multiple Points";
	calc.m_calc_type = P_CALC_POINT;

	// databases
	shared_ptr<Pan_Database> d_b = make_shared<Pan_Database>();
	d_b->m_file_name = db.first;
	d_b->m_type = db.second;
	calc.m_databases.push_back(pair<string, shared_ptr<Pan_Database>>(d_b->m_file_name, d_b));

	// units
	shared_ptr<Pan_Unit> unit_T = make_shared<Pan_Unit>();
	unit_T->m_name = "T";
	unit_T->m_unit = P_T_K;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_T->m_name, unit_T));

	shared_ptr<Pan_Unit> unit_P = make_shared<Pan_Unit>();
	unit_P->m_name = "P";
	unit_P->m_unit = P_P_atm;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_P->m_name, unit_P));

	// set unit for component amount
	shared_ptr<Pan_Unit> unit_m = make_shared<Pan_Unit>();
	unit_m->m_name = "n";
	unit_m->m_unit = P_M_mole;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_m->m_name, unit_m));

	// system
	calc.m_system.push_back(make_shared<Pan_System>());
	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];
	
	// components
	shared_ptr<Pan_Component> comp1 = make_shared<Pan_Component>();
	comp1->m_name = "Al";
	comp1->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp1);

	shared_ptr<Pan_Component> comp2 = make_shared<Pan_Component>();
	comp2->m_name = "Mg";
	comp2->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp2);

	shared_ptr<Pan_Component> comp3 = make_shared<Pan_Component>();
	comp3->m_name = "Zn";
	comp3->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp3);

	// calculation condition	
	sys_ptr->m_condition->m_physical_property = true;

	// phases
	shared_ptr<Pan_Phase> phase0 = make_shared<Pan_Phase>();
	phase0->m_name = "*";
	phase0->m_status = P_PHASE_SUSPENDED;
	sys_ptr->m_phases.push_back(phase0);

	shared_ptr<Pan_Phase> phase1 = make_shared<Pan_Phase>();
	phase1->m_name = "*";
	phase1->m_status = P_PHASE_ENTERED;
	sys_ptr->m_phases.push_back(phase1);

	// point
	shared_ptr<Pan_Point> pt1 = make_shared<Pan_Point>();
	shared_ptr<Pan_Statespace> st;
	st = make_shared<Pan_Statespace>();
	st->m_T = 800;
	st->m_P = 1;
	st->m_n.insert(pair<string, double> ("Al", 0.2));
	st->m_n.insert(pair<string, double> ("Mg", 0.3));
	st->m_n.insert(pair<string, double> ("Zn", 0.5));

	pt1->m_statespace = st;

	sys_ptr->m_points.push_back(pt1);
}

int PanEngineX_Test_2()
{
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
	cout << "|  Test Example 2 : Equilibria for Multiple Points  |\n";
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

	//=================================
	// Define a PanEngineX user pointer
	//=================================
	PanEngine* user;
	char msg[256];
	string s;

	cout << "Defining user pointer:" << endl;
	user = definePanEngineUser(msg);

	if (user == NULL) 
	{
		cout << "Error in defineUser" << endl;
		return 0;
	}


	//============================
	// Set PanEngine Configuration
	//============================
	// set "case_sensitive" before importing database
	cout << "Set PanEngine configuration:" << endl;
	map<string, string> config;
	config["case_sensitive"] = "true";
	s = user->pe_set_configuration(config);

	//==========================================
	// Load Thermodynamic Database in tdb Format
	//==========================================
	cout << "Reading database file:" << endl;

	pair<string, string> db = make_pair<string, string>("AlMgZn.tdb", "tdb");

	s = user->pe_import_database(db);

	if (!s.empty())
	{
		cout << "Error in open database [" << s << "]\n";
		return 0;
	}

	//==============================
	// Get Component and Phase Names
	//==============================
	vector<string> comp_name;
	vector<string> phase_name;
	s = user->pe_get_component_names(db, comp_name);
	s = user->pe_get_phase_names(db, phase_name);

	cout << "comp_name =\n";
	for_each(comp_name.begin(), comp_name.end(), [&](string c_name) {	cout << "\t" << c_name; }	);
	cout << endl;
	cout << "phase_name =\n";
	for_each(phase_name.begin(), phase_name.end(), [&](string p_name) {	cout << "\t" << p_name << endl; }	);
	cout << endl;

	//==============================
	// set up calculation conditions
	//==============================
	Pan_Calculation calc;
	set_calculation_condition_2(calc, db);
	shared_ptr<P_Point> calc_point = make_shared<P_Point>();

	s = user->pe_set_calculation_condition(calc);
		
	if (!s.empty()) {
		cout << "Error in calculation condition [" << s << "]\n";
		return 0;
	}

	//==========================
	// Initial Point Calculation
	//==========================
	// using pe_calc_point_global for first point
	s = user->pe_calc_point_global(calc_point);
	calc_point->print_debug(cout);


	//======================
	// Changing Compositions
	//======================
	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];
	int n_points = 0;
	for (double x_Mg = 0;  x_Mg <= 1; x_Mg += 0.1) {
		if (!valid_x(x_Mg)) continue;
		for (double x_Zn = 0;  x_Zn <= 1-x_Mg; x_Zn += 0.1) {
			if (!valid_x(x_Zn)) continue;
			double x_Al = 1 - x_Mg - x_Zn;
			if (!valid_x(x_Al)) continue;
			if (x_Al < 0) continue;

			++n_points;
			cout << "\n========= Point " << n_points << " =========\n";
			cout << "x[Al]=" << x_Al << "  x[Mg]=" << x_Mg << "  x[Zn]=" << x_Zn << endl;

			calc_point->m_st->m_comp["Al"]->m_x = x_Al;
			calc_point->m_st->m_comp["Mg"]->m_x = x_Mg;
			calc_point->m_st->m_comp["Zn"]->m_x = x_Zn;

			// using "pe_calc_point_global_with_initial_point" 
			// for calc_point with initial from the last calculation
			s = user->pe_calc_point_global_with_initial_point(calc_point);

			//calc_point->print_debug(cout);
		}
	}
	cout << "\nTotal number of points=" << n_points << endl;
	cout << "\n\n========= End of Example 2 =========\n\n";


	//===============================
	// Delete PanEngineX user pointer
	//===============================
	deletePanEngineUser(user);

	return 1;
}

