/////////////////////////////////////
//                                 //
//         PanEngine API           //
//                                 //
//         Version 2020            //
//                                 //
//        CompuTherm, LLC          //
//      Copyright, 2012-2020       //
//                                 //
//       www.computherm.com        //
//                                 //
/////////////////////////////////////

//==================================//
//                                  //
//    PanEngine Test Example 5      //
//                                  //
//    Solidification Simulation     //
//       for Multiple Alloys        //
//           (Al-Mg-Zn)             //
//                                  //
//==================================//


#include "std.h"
#include "PanEngineX.h"
#include "solidification.h"

#ifdef WIN32
#include <windows.h>
#endif
#include <sstream>


extern string output_result(string file_name, vector<Solidification_Node>& solidification_result);

bool progress_5(char* msg, Solidification_Parameter& solidification_param, vector<Solidification_Node>& node) {
	//cout << "T = " << node.rbegin()->pt.m_st->m_T << " K\n";
	return true;
}

void set_solidification_simulation_condition_5(Pan_Calculation& calc, pair<string, string> db, int alloy_number, bool is_scheil_model) {

	calc.m_name="Solidification Simulation for Multiple Alloys";
	calc.m_calc_type = P_CALC_SOLIDIFICATION;

	// databases
	shared_ptr<Pan_Database> d_b = make_shared<Pan_Database>();
	d_b->m_file_name = db.first;
	d_b->m_type = db.second;
	calc.m_databases.push_back(pair<string, shared_ptr<Pan_Database>>(d_b->m_file_name, d_b));

	// units
	shared_ptr<Pan_Unit> unit_T = make_shared<Pan_Unit>();
	unit_T->m_name = "T";
	unit_T->m_unit = P_T_K;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_T->m_name, unit_T));

	shared_ptr<Pan_Unit> unit_P = make_shared<Pan_Unit>();
	unit_P->m_name = "P";
	unit_P->m_unit = P_P_atm;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_P->m_name, unit_P));

	// set unit for component amount
	shared_ptr<Pan_Unit> unit_m = make_shared<Pan_Unit>();
	unit_m->m_name = "n";
	unit_m->m_unit = P_M_kg;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_m->m_name, unit_m));

	// system
	calc.m_system.push_back(make_shared<Pan_System>());
	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];

	// components
	vector <pair<string, double>> comp_name_amount;

	switch (alloy_number) {
	case 1 :
		comp_name_amount.push_back(pair<string, double>("Al", 70));
		comp_name_amount.push_back(pair<string, double>("Mg", 13));
		comp_name_amount.push_back(pair<string, double>("Zn", 17));
		break;
	case 2 :
		comp_name_amount.push_back(pair<string, double>("Al", 70));
		comp_name_amount.push_back(pair<string, double>("Mg", 10));
		comp_name_amount.push_back(pair<string, double>("Zn", 20));
		break;
	case 3 :
		comp_name_amount.push_back(pair<string, double>("Al", 90));
		comp_name_amount.push_back(pair<string, double>("Mg", 5));
		comp_name_amount.push_back(pair<string, double>("Zn", 5));
		break;
	default:
		comp_name_amount.push_back(pair<string, double>("Al", 20));
		comp_name_amount.push_back(pair<string, double>("Mg", 50));
		comp_name_amount.push_back(pair<string, double>("Zn", 30));
		break;
	}

	for (int i=0; i<(int)comp_name_amount.size(); ++i) {
		shared_ptr<Pan_Component> comp = make_shared<Pan_Component>();
		comp->m_name = comp_name_amount[i].first;
		comp->m_status = P_SELECTED;
		sys_ptr->m_components.push_back(comp);
	}

	// phases
	shared_ptr<Pan_Phase> phase0 = make_shared<Pan_Phase>();
	phase0->m_name = "*";
	phase0->m_status = P_PHASE_SUSPENDED;
	sys_ptr->m_phases.push_back(phase0);

	shared_ptr<Pan_Phase> phase1 = make_shared<Pan_Phase>();
	phase1->m_name = "*";
	phase1->m_status = P_PHASE_ENTERED;
	sys_ptr->m_phases.push_back(phase1);

	// starting point
	shared_ptr<Pan_Point> pt1 = make_shared<Pan_Point>();
	shared_ptr<Pan_Statespace> st;
	st = make_shared<Pan_Statespace>();
	st->m_T = 750+273;
	st->m_P = 1;

	for (int i=0; i<(int)comp_name_amount.size(); ++i) {
		st->m_n.insert(pair<string, double> (comp_name_amount[i].first, comp_name_amount[i].second));
	}

	pt1->m_statespace = st;

	sys_ptr->m_points.push_back(pt1);

	// calculation condition	
	if (is_scheil_model) 
	{
		sys_ptr->m_condition->m_model = "scheil";
	}
	else 
	{
		sys_ptr->m_condition->m_model = "lever";
	}
	
	sys_ptr->m_condition->m_physical_property = true;
	sys_ptr->m_condition->m_equilibriumType = "global";
	sys_ptr->m_condition->m_step_T_max = 20;
	sys_ptr->m_condition->m_step_T_min = 0.1;
	sys_ptr->m_condition->m_T_end = 300;
	sys_ptr->m_condition->m_liquid_amount_min = 0.005;
	sys_ptr->m_condition->m_liquid_lamda_min = 0.8;
	sys_ptr->m_condition->m_x_min = 1e-10;
}

string set_solidification_parameters_5(const Pan_Calculation& calc, Solidification_Parameter& s_param) {
	s_param.reset();

	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];
	string model=sys_ptr->m_condition->m_model;
	if (_stricmp(model.c_str(), "Scheil") == 0) {
		s_param.model_type = 0;
	} else
	if (_stricmp(model.c_str(), "lever") == 0) {
		s_param.model_type = 1;
	} else {
		s_param.model_type = -1;
	}

	s_param.m_solvent_name = string("Al");
	s_param.dT_min = sys_ptr->m_condition->m_step_T_min;
	s_param.dT_max = sys_ptr->m_condition->m_step_T_max;
	s_param.T_start = sys_ptr->m_points[0]->m_statespace->m_T;	// user given starting temperature
	s_param.T_end = sys_ptr->m_condition->m_T_end;
	s_param.T_range = fabs(s_param.T_start - s_param.T_end);
	if (s_param.T_range < 1) s_param.T_range = 1;

	s_param.dT_max = sys_ptr->m_condition->m_step_T_max;
	s_param.dT_min = sys_ptr->m_condition->m_step_T_min;

	s_param.min_liq_amount = sys_ptr->m_condition->m_liquid_amount_min;	// minimum liquidus amount to stop
	s_param.min_x = sys_ptr->m_condition->m_x_min;	// minimum amount of xj
	s_param.useAverageCompositionInSolidForScheil = true;

	return "";
}

int PanEngineX_Test_5()
{
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
	cout << "|  Test Example 5 : Solidification Simulation for Multiple Alloys  |\n";
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

	int alloy_number = 1;
	bool is_Scheil_model = true;

	//================================
	// Define a PanEngine user pointer
	//================================
	PanEngine* user;
	char msg[256];
	string s;

	cout << "Defining user pointer:" << endl;
	user = definePanEngineUser(msg);

	if (user == NULL)
	{
		cout << "Error in defineUser" << endl;
		return 0;
	}


	//============================
	// Set PanEngine Configuration
	//============================
	// set "case_sensitive" before importing database
	cout << "Set PanEngine configuration:" << endl;
	map<string, string> config;
	config["case_sensitive"] = "true";
	s = user->pe_set_configuration(config);

	//==========================================
	// Load Thermodynamic Database in tdb Format
	//==========================================
	cout << "Reading database file:" << endl;

	pair<string, string> db = make_pair<string, string>(string("AlMgZn.tdb"), string("tdb"));

	s = user->pe_import_database(db);

	if (!s.empty())
	{
		cout << "Error in open database [" << s << "]\n";
		return 0;
	}

	//==========================================
	// Input Solidification Simulation Condition
	//==========================================
	cout << "Input solidification simulation condition:" << endl;
	Pan_Calculation calc;
	Solidification_Parameter s_param;

	set_solidification_simulation_condition_5(calc, db, alloy_number, is_Scheil_model);
	set_solidification_parameters_5(calc, s_param);

	//======================================
	// Set Simulation Condition in PanEngine
	//======================================
	cout << "Set calculation condition:" << endl;
	s = user->pe_set_calculation_condition(calc);

	if (!s.empty()) {
		cout << "Error in calculation condition [" << s << "]\n";
		return 0;
	}

	//==========================
	// Solidification Simulating
	//==========================
	cout << "Solidification Simulating ..." << endl;
	vector <Solidification_Node> solidification_result;	// store solidification results
	string result;

	// set callback function
	s_param.progress_ptr = progress_5;

	result=user->pe_solidification_simulation(s_param, calc, solidification_result);

	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];
	for (double wMg=1; wMg<30; wMg += 2) {	
		double wAl = 70;
		double wZn = 100 - wMg - wAl;
		if (wAl < 0 || wAl > 100 || wMg < 0 || wMg > 100 || wZn < 0 || wZn > 100) {
			break;
		}

		sys_ptr->m_points[0]->m_statespace->m_n["Al"] = wAl;
		sys_ptr->m_points[0]->m_statespace->m_n["Mg"] = wMg;
		sys_ptr->m_points[0]->m_statespace->m_n["Zn"] = wZn;

		s_param.reset();
		solidification_result.clear();
		s = user->pe_set_calculation_condition(calc);
		cout << "w(Al) = " << wAl << "%  " 
			 << "w(Mg) = " << wMg << "%  " 
			 << "w(Zn) = " << wZn << "%\n";
		result=user->pe_solidification_simulation(s_param, calc, solidification_result);
	}

//	output_result("my_test_5.dat", solidification_result);

	cout << "\n\n========= End of Example 5 =========\n\n";


	//==============================
	// Delete PanEngine user pointer
	//==============================
	deletePanEngineUser(user);

	return 1;
}