/////////////////////////////////////
//                                 //
//         PanEngine API           //
//                                 //
//         Version 2020            //
//                                 //
//        CompuTherm, LLC          //
//      Copyright, 2012-2020       //
//                                 //
//       www.computherm.com        //
//                                 //
/////////////////////////////////////

//==================================//
//                                  //
//    PanEngine Test Example 7      //
//                                  //
//      Lowest Liquidus Point       //
//        (Al-Mg-Zn System)         //
//----------------------------------//
//  Use three stages to find the    //
//  lowest liquidus point           //
//                                  //
// Stage 1                          //
//    Search whole compositional    //
//    range with dx = 1.0 / n_steps //
// Stage 2                          //
//    Search locally with           //
//    dx = 1.0 / n_steps^2          //
// Stage 3                          //
//    Search Locally with           //
//    dx = 1.0 / n_steps^3          //
//                                  //
//==================================//


#include "std.h"
#include "PanEngineX.h"

extern bool valid_x(double& x);


void set_calculation_condition_7(Pan_Calculation& calc, pair<string, string> db) {

	calc.m_name="Lowest Liquidus Point";
	calc.m_calc_type = P_CALC_POINT;
	calc.m_calc_type = P_CALC_SOLIDIFICATION;

	// databases
	shared_ptr<Pan_Database> d_b = make_shared<Pan_Database>();
	d_b->m_file_name = db.first;
	d_b->m_type = db.second;
	calc.m_databases.push_back(pair<string, shared_ptr<Pan_Database>>(d_b->m_file_name, d_b));

	// units
	shared_ptr<Pan_Unit> unit_T = make_shared<Pan_Unit>();
	unit_T->m_name = "T";
	unit_T->m_unit = P_T_K;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_T->m_name, unit_T));

	shared_ptr<Pan_Unit> unit_P = make_shared<Pan_Unit>();
	unit_P->m_name = "P";
	unit_P->m_unit = P_P_atm;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_P->m_name, unit_P));

	// set unit for component amount
	shared_ptr<Pan_Unit> unit_m = make_shared<Pan_Unit>();
	unit_m->m_name = "n";
	unit_m->m_unit = P_M_mole;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_m->m_name, unit_m));

	// system
	calc.m_system.push_back(make_shared<Pan_System>());
	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];
	
	// components
	shared_ptr<Pan_Component> comp1 = make_shared<Pan_Component>();
	comp1->m_name = "Al";
	comp1->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp1);

	shared_ptr<Pan_Component> comp2 = make_shared<Pan_Component>();
	comp2->m_name = "Mg";
	comp2->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp2);

	shared_ptr<Pan_Component> comp3 = make_shared<Pan_Component>();
	comp3->m_name = "Zn";
	comp3->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp3);

	// phases
	shared_ptr<Pan_Phase> phase0 = make_shared<Pan_Phase>();
	phase0->m_name = "*";
	phase0->m_status = P_PHASE_ENTERED;
	sys_ptr->m_phases.push_back(phase0);

	// point
	shared_ptr<Pan_Point> pt1 = make_shared<Pan_Point>();
	shared_ptr<Pan_Statespace> st;
	st = make_shared<Pan_Statespace>();
	st->m_T = 600;
	st->m_P = 1;
	st->m_n.insert(pair<string, double> ("Al", 0.2));
	st->m_n.insert(pair<string, double> ("Mg", 0.3));
	st->m_n.insert(pair<string, double> ("Zn", 0.5));

	pt1->m_statespace = st;

	sys_ptr->m_points.push_back(pt1);
}

int PanEngineX_Test_7()
{
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
	cout << "|  Test Example 7 : Lowest Liquidus Point  |\n";
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

	//================================
	// Define a PanEngine user pointer
	//================================
	PanEngine* user;
	char msg[256];
	string s;

	cout << "Defining user pointer:" << endl;
	user = definePanEngineUser(msg);

	if (user == NULL) 
	{
		cout << "Error in defineUser" << endl;
		return 0;
	}

	//============================
	// Set PanEngine Configuration
	//============================
	// set "case_sensitive" before importing database
	cout << "Set PanEngine configuration:" << endl;
	map<string, string> config;
	config["case_sensitive"] = "true";
	s = user->pe_set_configuration(config);

	//==========================================
	// Load Thermodynamic Database in tdb Format
	//==========================================
	cout << "Reading database file:" << endl;

	pair<string, string> db = make_pair<string, string>("AlMgZn.tdb", "tdb");

	s = user->pe_import_database(db);

	if (!s.empty())
	{
		cout << "Error in open database [" << s << "]\n";
		return 0;
	}

	//==============================
	// Get Component and Phase Names
	//==============================
	vector<string> comp_name;
	vector<string> phase_name;
	s = user->pe_get_component_names(db, comp_name);
	s = user->pe_get_phase_names(db, phase_name);
	
	cout << "comp_name =\n";
	for_each(comp_name.begin(), comp_name.end(), [&](string c_name) {	cout << "\t" << c_name; }	);
	cout << endl;
	cout << "phase_name =\n";
	for_each(phase_name.begin(), phase_name.end(), [&](string p_name) {	cout << "\t" << p_name << endl; }	);
	cout << endl;

	//==============================
	// set up calculation conditions
	//==============================
	Pan_Calculation calc;
	set_calculation_condition_7(calc, db);
	shared_ptr<P_Point> calc_point = make_shared<P_Point>();

	s = user->pe_set_calculation_condition(calc);

	if (!s.empty()) {
		cout << "Error in calculation condition [" << s << "]\n";
		return 0;
	}

	//======================
	// Find liquidus surface
	//======================
	shared_ptr<P_Point> liquidus_point = make_shared<P_Point>();

	string liquid_name =string("Liquid");

	s=user->pe_find_liquidus_surface(liquid_name, calc_point);
  
	if (!s.empty()) {
		cout << s << endl; 
		return 0;
	}
	
	cout << "\n\n=============================================";	
	calc_point->print_debug(cout);
	cout << endl;

	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];
	shared_ptr<P_Point> calc_point_min_T = make_shared<P_Point>();
	//==============================================================
	// Stage 1: Search whole compositional range, dx = 1.0 / n_steps
	//==============================================================
	double Tmin = 10000;
	double dx;
	int n_steps = 10;
	double x_Al, x_Mg, x_Zn;
	dx = 1.0 / n_steps;

	for (int j = 0; j <= n_steps; ++j) {
		x_Mg = dx * j;
		if (!valid_x(x_Mg)) continue;
		for (int k = 0; k <= n_steps-j; ++k) {
			x_Zn = dx * k;
			if (!valid_x(x_Zn)) continue;
			x_Al = 1 - x_Mg - x_Zn;
			if (!valid_x(x_Al)) continue;

			calc_point->m_st->m_comp["Al"]->m_x = x_Al;
			calc_point->m_st->m_comp["Mg"]->m_x = x_Mg;
			calc_point->m_st->m_comp["Zn"]->m_x = x_Zn;

			// liquidus surface
			s = user->pe_find_liquidus_surface(liquid_name, calc_point);
			//cout << "s=" << s << endl;
			//cout << "\n\n=============================================";
			//calc_point->print_debug(cout);
			//cout << endl;

			if (calc_point->m_st->m_T<Tmin) {
				*calc_point_min_T = *calc_point;
				Tmin = calc_point->m_st->m_T;
			}			
			
			cout << "x[Al]=" << x_Al << "  x[Mg]=" << x_Mg << "  x[Zn]=" << x_Zn 
				<< "  T=" << calc_point->m_st->m_T << " K ";
			for_each(calc_point->m_ppt.begin(), calc_point->m_ppt.end(), [&](shared_ptr<P_Phase_Point> ppt) { cout << " " << ppt->m_phase_name; }	);
			cout << "  Tmin=" << Tmin << " K" << endl;
		}
	}

	cout << "\nAfter Stage 1\n";
	cout << "Tmin=" << Tmin << endl;
//	cout << "calc_point_min_TL\n";
//	calc_point_min_T->print_debug(cout);
	cout << endl;

	//===============================================
	// Stage 2:  Search Locally, dx = 1.0 / n_steps^2
	//===============================================
	*calc_point = *calc_point_min_T;
	dx /= n_steps;
	double x_Al_0, x_Mg_0, x_Zn_0;

	x_Al_0 = calc_point_min_T->m_st->m_comp["Al"]->m_x;
	x_Mg_0 = calc_point_min_T->m_st->m_comp["Mg"]->m_x;
	x_Zn_0 = calc_point_min_T->m_st->m_comp["Zn"]->m_x;

	for (int j = -n_steps/2; j <= n_steps/2; ++j) {
		x_Mg = x_Mg_0 + dx * j;
		if (!valid_x(x_Mg)) continue;
		for (int k = -n_steps/2; k <= n_steps/2; ++k) {
			x_Zn = x_Zn_0 + dx * k;
			if (!valid_x(x_Zn)) continue;
			x_Al = 1 - x_Mg - x_Zn;
			if (!valid_x(x_Al)) continue;

			calc_point->m_st->m_comp["Al"]->m_x = x_Al;
			calc_point->m_st->m_comp["Mg"]->m_x = x_Mg;
			calc_point->m_st->m_comp["Zn"]->m_x = x_Zn;

			// liquidus surface
			s=user->pe_find_liquidus_surface(liquid_name, calc_point);
			if (calc_point->m_st->m_T<Tmin) {
				*calc_point_min_T = *calc_point;
				Tmin = calc_point->m_st->m_T;
			}			
			
			cout << "x[Al]=" << x_Al << "  x[Mg]=" << x_Mg << "  x[Zn]=" << x_Zn 
				<< "  T=" << calc_point->m_st->m_T << " K ";
			for_each(calc_point->m_ppt.begin(), calc_point->m_ppt.end(), [&](shared_ptr<P_Phase_Point> ppt) { cout << " " << ppt->m_phase_name; }	);
			cout << "  Tmin=" << Tmin << " K" << endl;
		}
	}

	cout << "\nAfter Stage 2\n";
	cout << "Tmin=" << Tmin << endl;
//	cout << "calc_point_min_TL\n";
//	calc_point_min_T->print_debug(cout);
	cout << endl;

	//===============================================
	// Stage 3:  Search Locally, dx = 1.0 / n_steps^3
	//===============================================
	*calc_point = *calc_point_min_T;
	dx /= n_steps;

	x_Al_0 = calc_point_min_T->m_st->m_comp["Al"]->m_x;
	x_Mg_0 = calc_point_min_T->m_st->m_comp["Mg"]->m_x;
	x_Zn_0 = calc_point_min_T->m_st->m_comp["Zn"]->m_x;

	for (int j = -n_steps/2; j <= n_steps/2; ++j) {
		x_Mg = x_Mg_0 + dx * j;
		if (!valid_x(x_Mg)) continue;
		for (int k = -n_steps/2; k <= n_steps/2; ++k) {
			x_Zn = x_Zn_0 + dx * k;
			if (!valid_x(x_Zn)) continue;
			x_Al = 1 - x_Mg - x_Zn;
			if (!valid_x(x_Al)) continue;

			calc_point->m_st->m_comp["Al"]->m_x = x_Al;
			calc_point->m_st->m_comp["Mg"]->m_x = x_Mg;
			calc_point->m_st->m_comp["Zn"]->m_x = x_Zn;

			// liquidus surface
			s=user->pe_find_liquidus_surface(liquid_name, calc_point);
			if (calc_point->m_st->m_T<Tmin) {
				*calc_point_min_T = *calc_point;
				Tmin = calc_point->m_st->m_T;
			}			
			
			cout << "x[Al]=" << x_Al << "  x[Mg]=" << x_Mg << "  x[Zn]=" << x_Zn 
				<< "  T=" << calc_point->m_st->m_T << " K ";
			for_each(calc_point->m_ppt.begin(), calc_point->m_ppt.end(), [&](shared_ptr<P_Phase_Point> ppt) { cout << " " << ppt->m_phase_name; }	);
			cout << "  Tmin=" << Tmin << " K" << endl;
		}
	}

	cout << "\nAfter Stage 3\n";
	cout << "Tmin=" << Tmin << " K" << endl;
	cout << "calc_point_min_TL\n";
	calc_point_min_T->print_debug(cout);
	cout << endl;

	cout << "\n\n========= End of Example 7 =========\n\n";


	//===============================
	// Delete PanEngineX user pointer
	//===============================
	deletePanEngineUser(user);

	return 1;
}

