/////////////////////////////////////
//                                 //
//         PanEngine API           //
//                                 //
//         Version 2020            //
//                                 //
//        CompuTherm, LLC          //
//      Copyright, 2012-2020       //
//                                 //
//       www.computherm.com        //
//                                 //
/////////////////////////////////////

//==================================//
//                                  //
//    PanEngine Test Example 8      //
//                                  //
//       Kinetic Properties         //
//       and Hessian Matrix         //
//----------------------------------//
//  Define a PanEngine user pointer //
//  Set PanEngine configuration     //
//  Read a database                 //
//  Define a subsystem              //
//  Set calculation condition       //
//  Get kinetic properties          //
//  Get eigenvalues, eigenvectors   //
//      and determinant of Hessian  //
//      of Gibbs free energy        //
//  Delete a PanEngine user pointer //
//                                  //
//==================================//


#include "std.h"
#include "PanEngineX.h"


//===================================
//    PanEngine Test Example 8
//
//   Kinetic Properties
//
//-----------------------------------
//  Define a PanEngine user pointer
//  Set PanEngine configuration
//  Read a database
//  Define a subsystem
//  Set calculation condition
//  Get Kinetic Properties
//  Get eigenvalues, eigenvectors and 
//      determinant of Hessian of Gibbs free energy
//  Delete a PanEngine user pointer
//
//===================================

void set_calculation_condition_8(Pan_Calculation& calc, pair<string, string> db) {

	calc.m_name="Kinetic Properties";
	calc.m_calc_type = P_CALC_POINT;

	// databases
	shared_ptr<Pan_Database> d_b = make_shared<Pan_Database>();
	d_b->m_file_name = db.first;
	d_b->m_type = db.second;
	calc.m_databases.push_back(pair<string, shared_ptr<Pan_Database>>(d_b->m_file_name, d_b));

	// units
	shared_ptr<Pan_Unit> unit_T = make_shared<Pan_Unit>();
	unit_T->m_name = "T";
	unit_T->m_unit = P_T_K;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_T->m_name, unit_T));

	shared_ptr<Pan_Unit> unit_P = make_shared<Pan_Unit>();
	unit_P->m_name = "P";
	unit_P->m_unit = P_P_atm;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_P->m_name, unit_P));

	// set unit for component amount
	shared_ptr<Pan_Unit> unit_m = make_shared<Pan_Unit>();
	unit_m->m_name = "n";
	unit_m->m_unit = P_M_mole;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>> (unit_m->m_name, unit_m));

	// system
	calc.m_system.push_back(make_shared<Pan_System>());
	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];
	
	// components
	shared_ptr<Pan_Component> comp1 = make_shared<Pan_Component>();
	comp1->m_name = "Fe";
	comp1->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp1);

	shared_ptr<Pan_Component> comp2 = make_shared<Pan_Component>();
	comp2->m_name = "Ni";
	comp2->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp2);

	shared_ptr<Pan_Component> comp3 = make_shared<Pan_Component>();
	comp3->m_name = "Cr";
	comp3->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp3);

	// phases
	shared_ptr<Pan_Phase> phase0 = make_shared<Pan_Phase>();
	phase0->m_name = "*";
	phase0->m_status = P_PHASE_SUSPENDED;
	sys_ptr->m_phases.push_back(phase0);

	shared_ptr<Pan_Phase> phase1 = make_shared<Pan_Phase>();
	phase1->m_name = "FCC_A1";
	phase1->m_status = P_PHASE_ENTERED;
	sys_ptr->m_phases.push_back(phase1);


	// calculation condition	
	// calculate driving forces for all unsuspended phases
	sys_ptr->m_condition->m_driving_force = false;
	sys_ptr->m_condition->m_kinetic_property = true;

	sys_ptr->m_condition->m_solvent = string("Ni");

	// point
	shared_ptr<Pan_Point> pt1 = make_shared<Pan_Point>();
	shared_ptr<Pan_Statespace> st;
	st = make_shared<Pan_Statespace>();
	st->m_T = 1000;
	st->m_P = 1;
	st->m_n.insert(pair<string, double> ("Fe", 0.2));
	st->m_n.insert(pair<string, double> ("Ni", 0.3));
	st->m_n.insert(pair<string, double> ("Cr", 0.5));

	pt1->m_statespace = st;

	sys_ptr->m_points.push_back(pt1);
}

int PanEngineX_Test_8()
{
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
	cout << "|  Test Example 8 : Kinetic Properties  |\n";
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

	//=================================
	// Define a PanEngineX user pointer
	//=================================
	PanEngine* user;
	char msg[256];
	string s;

	cout << "Defining user pointer:" << endl;
	user = definePanEngineUser(msg);

	if (user == NULL) 
	{
		cout << "Error in defineUser" << endl;
		return 0;
	}


	//============================
	// Set PanEngine Configuration
	//============================
	// set "case_sensitive" before importing database
	cout << "Set PanEngine configuration:" << endl;
	map<string, string> config;
	config["case_sensitive"] = "true";
	s = user->pe_set_configuration(config);

	//==========================================
	// Load Thermodynamic Database in tdb Format
	//==========================================
	cout << "Reading database file:" << endl;

	pair<string, string> db = make_pair<string, string>("FeNiCr.tdb", "tdb");

	s = user->pe_import_database(db);

	if (!s.empty())
	{
		cout << "Error in open database [" << s << "]\n";
		return 0;
	}

	//==============================
	// Get Component and Phase Names
	//==============================
	vector<string> comp_name;
	vector<string> phase_name;
	s = user->pe_get_component_names(db, comp_name);
	s = user->pe_get_phase_names(db, phase_name);

	cout << "comp_name =\n";
	for_each(comp_name.begin(), comp_name.end(), [&](string c_name) {	cout << "\t" << c_name; }	);
	cout << endl;
	cout << "phase_name =\n";
	for_each(phase_name.begin(), phase_name.end(), [&](string p_name) {	cout << "\t" << p_name << endl; }	);
	cout << endl;

	//==============================
	// set up calculation conditions
	//==============================
	Pan_Calculation calc;
	set_calculation_condition_8(calc, db);
	shared_ptr<P_Point> calc_point = make_shared<P_Point>();

	s = user->pe_set_calculation_condition(calc);
	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];

	if (!s.empty()) {
		cout << "Error in calculation condition [" << s << "]\n";
		return 0;
	}

	//==================
	// Point Calculation
	//==================
	cout << "\n\n========= Point 1 =========\n";
	s = user->pe_calc_point_global(calc_point);
	cout << "after pe_calc_point_global  s=[" << s << "]\n";

	calc_point->print_debug(cout);


	cout << "\n\n========= Point 2 =========\n";
	calc_point->m_st->m_T -= 100.0;
	s = user->pe_calc_point_global_with_initial_point(calc_point);
	calc_point->print_debug(cout);

	if (calc_point->m_ppt[0]->m_thermodynamic.first == true) {
		auto K_Prop = calc_point->m_ppt[0]->m_thermodynamic.second;

		if (K_Prop.exist("ThF")) {
			cout << "Thermodynamic Factors:" << endl;
			Val v;
			for (auto c = calc_point->m_st->m_comp.begin(); c != calc_point->m_st->m_comp.end(); ++c) {
				v.m_comp = (*c).first;
				for (auto gc = calc_point->m_st->m_comp.begin(); gc != calc_point->m_st->m_comp.end(); ++gc) {
					v.m_grad_comp = (*gc).first;
					bool exist = K_Prop.get_value("ThF", v);
					if (exist) {
						cout << "    ThF[" << v.m_comp << "," << v.m_grad_comp << "] = " << v.m_val << endl;
					}
				}
			}
		}

		cout << endl << endl;
	}

	if (calc_point->m_ppt[0]->m_kinetic.first == true) {
		cout << "Kinetic Properties:" << endl;
		auto K_Prop = calc_point->m_ppt[0]->m_kinetic.second;

		if (K_Prop.exist("M")) {
			cout << "  Mobility:" << endl;		
			Val v;
			for (auto c = calc_point->m_st->m_comp.begin(); c != calc_point->m_st->m_comp.end(); ++c) {
				v.m_comp = (*c).first;
				bool exist = K_Prop.get_value("M", v);
				if (exist) {
					cout << "    M[" << v.m_comp << "] = " << v.m_val << endl;
				}
			}
		}

		if (K_Prop.exist("DC")) {
			cout << "  Diffusivity:" << endl;
			Val v;
			v.m_ref_comp = sys_ptr->m_condition->m_solvent;
			for (auto c = calc_point->m_st->m_comp.begin(); c != calc_point->m_st->m_comp.end(); ++c) {
				v.m_comp = (*c).first;
				for (auto gc = calc_point->m_st->m_comp.begin(); gc != calc_point->m_st->m_comp.end(); ++gc) {
					v.m_grad_comp = (*gc).first;
					bool exist = K_Prop.get_value("DC", v);
					if (exist) {
						cout << "    DC[" << v.m_comp << "," << v.m_grad_comp << ":" << v.m_ref_comp << "] = " << v.m_val << endl;
					}
				}
			}
		}
		cout << endl << endl;
	}

	if (calc_point->m_ppt[0]->m_Hess.first == true) {
		cout << "Hessian Matrix of G:" << endl;
		cout << "  HSN = " << endl;
		auto HSN_MTX = calc_point->m_ppt[0]->m_Hess.second.m_Hessian;
		for (auto i = HSN_MTX.begin(); i != HSN_MTX.end(); ++i) {
			for (auto j = (*i).begin(); j != (*i).end(); ++j) {
				cout << "    <" << get<0>((*j)) << ", " << get<1>((*j)) << ", " << get<2>((*j)) << ">  ";
			}
			cout << endl;
		}
		cout << endl;

		cout << "  Determinant of Hessian Matrix of G:" << endl;
		cout << "    |HSN| = " << calc_point->m_ppt[0]->m_Hess.second.m_HSN << endl;
		cout << endl;

		cout << "  Eigenvalues and Eigenvectors of Hessian Matrix of G:" << endl;
		auto eigen = calc_point->m_ppt[0]->m_Hess.second.m_eigen;
		for (auto it = eigen.begin(); it != eigen.end(); ++it) {
			cout << "    eVal = " << (*it).first << "  eVec:";
			for (auto v = (*it).second.begin(); v != (*it).second.end(); ++v) {
				cout << " <" << (*v).first << ", " << (*v).second << ">";
			}
			cout << endl;
		}
		cout << endl;
	}


	cout << "\n\n========= End of Example 8 =========\n\n";


	//===============================
	// Delete PanEngineX user pointer
	//===============================
	deletePanEngineUser(user);

	return 1;
}

