/////////////////////////////////////
//                                 //
//         PanEngine API           //
//                                 //
//         Version 2020            //
//                                 //
//        CompuTherm, LLC          //
//      Copyright, 2012-2020       //
//                                 //
//       www.computherm.com        //
//                                 //
/////////////////////////////////////


//==================================//
//                                  //
//    PanEngine Test Example 9      //
//                                  // 
//  Parallel Tangent Equilibrium    //
//           (Ni-Al-Nb)             //
//----------------------------------//
//  First calculate global          //  
//  equilibrium. Then, use the      // 
//  result as initial to calculate  // 
//  parallel tangent equilibrium    // 
//                                  //
//==================================//


#include "std.h"
#include "PanEngineX.h"


extern long long counter();
extern long long frequency();
extern string format_time(const long long start, const long long finish);


void set_calculation_condition_9(Pan_Calculation& calc, pair<string, string> db) {

	calc.m_name = "Parallel Tangent for NiAlNb";
	calc.m_calc_type = P_CALC_POINT;

	// databases
	shared_ptr<Pan_Database> d_b = make_shared<Pan_Database>();
	d_b->m_file_name = db.first;
	d_b->m_type = db.second;
	calc.m_databases.push_back(pair<string, shared_ptr<Pan_Database>>(d_b->m_file_name, d_b));

	// units
	shared_ptr<Pan_Unit> unit_T = make_shared<Pan_Unit>();
	unit_T->m_name = "T";
	unit_T->m_unit = P_T_K;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>>(unit_T->m_name, unit_T));

	shared_ptr<Pan_Unit> unit_P = make_shared<Pan_Unit>();
	unit_P->m_name = "P";
	unit_P->m_unit = P_P_atm;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>>(unit_P->m_name, unit_P));

	// set unit for component amount
	shared_ptr<Pan_Unit> unit_m = make_shared<Pan_Unit>();
	unit_m->m_name = "n";
	unit_m->m_unit = P_M_mole;
	calc.m_units.insert(pair<string, shared_ptr<Pan_Unit>>(unit_m->m_name, unit_m));

	// system
	calc.m_system.push_back(make_shared<Pan_System>());
	shared_ptr<Pan_System> sys_ptr = calc.m_system[0];

	// components
	shared_ptr<Pan_Component> comp1 = make_shared<Pan_Component>();
	comp1->m_name = "Ni";
	comp1->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp1);

	shared_ptr<Pan_Component> comp2 = make_shared<Pan_Component>();
	comp2->m_name = "Al";
	comp2->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp2);

	shared_ptr<Pan_Component> comp3 = make_shared<Pan_Component>();
	comp3->m_name = "Nb";
	comp3->m_status = P_SELECTED;
	sys_ptr->m_components.push_back(comp3);

	// calculation condition	
	sys_ptr->m_condition->m_physical_property = true;
	//	sys_ptr->m_condition->m_thermo_property = true;

	// phases
	shared_ptr<Pan_Phase> phase0 = make_shared<Pan_Phase>();
	phase0->m_name = "*";
	phase0->m_status = P_PHASE_SUSPENDED;
	sys_ptr->m_phases.push_back(phase0);

	shared_ptr<Pan_Phase> phase1 = make_shared<Pan_Phase>();
	phase1->m_name = "Fcc";
	phase1->m_status = P_PHASE_ENTERED;
	sys_ptr->m_phases.push_back(phase1);

	shared_ptr<Pan_Phase> phase2 = make_shared<Pan_Phase>();
	phase2->m_name = "gamma_double_prime";
	phase2->m_status = P_PHASE_ENTERED;
	sys_ptr->m_phases.push_back(phase2);

	// define "Ni" as the solvent component 
	// so that the independent variables will be x(Al) and x(Nb)
	// for the Hessian matrix
	sys_ptr->m_condition->m_solvent = string("Ni");
	
	// point
	shared_ptr<Pan_Point> pt1 = make_shared<Pan_Point>();
	shared_ptr<Pan_Statespace> st;
	st = make_shared<Pan_Statespace>();
	st->m_T = 1070;
	st->m_P = 1;
	st->m_n.insert(pair<string, double>("Ni", 0.94));
	st->m_n.insert(pair<string, double>("Al", 0.02));
	st->m_n.insert(pair<string, double>("Nb", 0.04));

	pt1->m_statespace = st;

	sys_ptr->m_points.push_back(pt1);
}

int PanEngineX_Test_9()
{
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
	cout << "|  Test Example 9 : Parallel Tangent for Phase Field  |\n";
	cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
	//=================================
	// Define a PanEngineX user pointer
	//=================================
	PanEngine* user;
	char msg[256];
	string s;

	cout << "Defining user pointer:" << endl;
	user = definePanEngineUser(msg);

	if (user == NULL)
	{
		cout << "Error in defineUser" << endl;
		return 0;
	}


	//============================
	// Set PanEngine Configuration
	//============================
	// set "case_sensitive" before importing database
	cout << "Set PanEngine configuration:" << endl;
	map<string, string> config;
	config["case_sensitive"] = "true";
	s = user->pe_set_configuration(config);

	//==========================================
	// Load Thermodynamic Database in tdb Format
	//==========================================
	cout << "Reading database file:" << endl;

	pair<string, string> db = pair<string, string>("NiAlNb_Pseudo.tdb", "tdb");

	s = user->pe_import_database(db);

	if (!s.empty())
	{
		cout << "Error in open database [" << s << "]\n";
		return 0;
	}

	//==============================
	// Get Component and Phase Names
	//==============================
	vector<string> comp_name;
	vector<string> phase_name;
	s = user->pe_get_component_names(db, comp_name);
	s = user->pe_get_phase_names(db, phase_name);

	cout << "comp_name =\n";
	for_each(comp_name.begin(), comp_name.end(), [&](string c_name) {	cout << "\t" << c_name; });
	cout << endl;
	cout << "phase_name =\n";
	for_each(phase_name.begin(), phase_name.end(), [&](string p_name) {	cout << "\t" << p_name << endl; });
	cout << endl;

	//==============================
	// set up calculation conditions
	//==============================
	Pan_Calculation calc;
	set_calculation_condition_9(calc, db);
	shared_ptr<P_Point> calc_point = make_shared<P_Point>();

	s = user->pe_set_calculation_condition(calc);

	if (!s.empty()) {
		cout << "Error in calculation condition [" << s << "]\n";
		return 0;
	}

	//==========================
	// Initial Point Calculation
	//==========================
	// using pe_calc_point_global for first point
	s = user->pe_calc_point_global(calc_point);
	cout << "Common tagent equilibrium:" << endl;
	calc_point->print_debug(cout);

	//============================
	// Parallel tagent equilibrium
	//============================
	calc_point->m_ppt[0]->m_f = 0.5;
	calc_point->m_ppt[1]->m_f = 0.5;

	s = user->pe_calc_point_local_with_initial_point(calc_point, true);
	cout << "Parallel tagent equilibrium from PanEngine:" << endl;
	calc_point->print_debug(cout);


	if (s.empty() == false)
	{
		cout << s << endl;
		return 0;
	}

	long long start_t, start_t0;
	long long end_t;


	random_device rd;   // non-deterministic generator  
	mt19937 gen(rd());  // to seed mersenne twister.  
	uniform_int_distribution<> r_T(1050, 1080);
	uniform_int_distribution<> r_x_Al(20, 30);	// x_Al * 1000
	uniform_int_distribution<> r_x_Nb(30, 50);	// x_Nb * 1000
	uniform_int_distribution<> r_f_Fcc(20, 70);	// f_Fcc * 100
																							
	// reset time
	start_t = counter();
	start_t0 = counter();

	//int n_loops = 100;
	int n_loops = 10;
	int n_max_points = 1000;

	for (int k = 0; k < n_loops; ++k) {
		for (int i = 0; i < n_max_points; ++i) {
			//		cout << "\n\n=======   Point #" << i << "   ============" << endl;
			double T = r_T(gen);
			double x_Al = 0.001 * r_x_Al(gen);
			double x_Nb = 0.001 * r_x_Nb(gen);
			double x_Ni = 1 - x_Al - x_Nb;
			double f_Fcc = 0.01 * r_f_Fcc(gen);

			calc_point->set_T(T);
			calc_point->m_st->m_comp["Al"]->m_x = x_Al;
			calc_point->m_st->m_comp["Nb"]->m_x = x_Nb;
			calc_point->m_st->m_comp["Ni"]->m_x = x_Ni;

			for (auto & ppt : calc_point->m_ppt) {
				if (ppt->m_phase_name == "Fcc") {
					ppt->m_f = f_Fcc;
				}
				else if (ppt->m_phase_name == "gamma_double_prime") {
					ppt->m_f = 1 - f_Fcc;
				}

			}

			s = user->pe_calc_point_local_with_initial_point(calc_point, true);
			cout << "@A Parallel tagent equilibrium from PanEngine:" << endl;
			calc_point->print_debug(cout);
		}

		end_t = counter();

		string t = format_time(start_t, end_t);

	//	double time = (double)(end_t - start_t) / (double)frequency();
		cout << k << "\t" << t << endl;
		start_t = counter();
	}

	cout << endl;

	end_t = counter();
	string total_time = format_time(start_t0, end_t);
	double time_per_point_1 = (double)(end_t - start_t0) / ((double)frequency() * n_loops * n_max_points);


	cout << "Total Time = " << total_time << endl;
	cout << "\nTotal number of points = " << n_loops * n_max_points << endl;
	cout << "Time/Point = " << time_per_point_1 << " seconds" << endl << endl;

	cout << "\n\n========= End of Example 9 =========\n\n";

	return 1;
}
