/////////////////////////////////////
//                                 //
//         PanEngine API           //
//                                 //
//         Version 2020            //
//        CompuTherm, LLC          //
//       Copyright, 2012-2020      //
//                                 //
//        www.computherm.com       //
//                                 //
/////////////////////////////////////

#pragma once

#include "stl.h"
#include "Pan_Global.h"
#include "solidification.h"

#ifdef __GNUC__
#define PANENGINE_API __attribute__((visibility("default"))) 
#else
#define PANENGINE_API _declspec(dllexport)
#endif


class PANENGINE_API PanEngine {
public:
	PanEngine(void);

	virtual string pe_set_configuration(map<string, string>& config) = 0;
	// If config["case_sensitive"] = "true", database file will be read as it is. The case for all names of components,
	// species and phases used in calculation must be the same as defined in database file.

	// If config["case_sensitive"] = "false", database file will be converted to all capital letters in PanEngine.
	// All names of components, species and phases used in calculation must be capital letters.


	virtual string pe_import_database(pair<string, string>& db, bool append = false, const pair<string, string>& db_to = pair<string, string>()) = 0;
	virtual string pe_import_database(string& db, const string& db_to = string("")) = 0;
	virtual string pe_import_database(vector<pair<string, string>>& db) = 0;

	// import a database into PanEngine

	virtual string pe_export_subsystem_database(pair<string, string>& db, const string subsystem_name = string("sub.tdb"), const vector<string>& comp_name = vector<string>()) = 0;
	virtual string pe_export_subsystem_database(string& db, const string subsystem_name = string("sub.tdb"), const vector<string>& comp_name = vector<string>()) = 0;
	// only work for tdb type of database. if comp_name is empty, use current selected component set.

	virtual string pe_get_database_version(pair<string, string>& name_version) = 0;
	// name_version.first = name of current database;  name_version.second = version of current database

	virtual string pe_get_panengine_version(string& version) = 0;

	virtual string pe_get_component_names(pair<string, string>& db, vector<string>& comp_name) = 0;
	virtual string pe_get_component_names(string& db, vector<string>& comp_name) = 0;	
	virtual string pe_get_phase_names(pair<string, string>& db, vector<string>& phase_name) = 0;
	virtual string pe_get_phase_names(string& db, vector<string>& phase_name) = 0;
	virtual string pe_get_phase_names(pair<string, string>& db, vector<string>& phase_name, bool& has_pressure, vector<string> comp_name = vector<string>()) = 0;
	virtual string pe_get_phase_names(string& db, vector<string>& phase_name, bool& has_pressure, vector<string> comp_name = vector<string>()) = 0;
	virtual string pe_get_all_phase_names(pair<string, string>& db, vector<string>& phase_name, bool& has_pressure, vector<string> comp_name = vector<string>()) = 0;
	virtual string pe_get_all_phase_names(string& db, vector<string>& phase_name, bool& has_pressure, vector<string> comp_name = vector<string>()) = 0;
	virtual string pe_get_phase_status(pair<string, string>& db, vector<pair<string, PAN_PHASE_STATUS>>& phase_name_status) = 0;
	virtual string pe_get_phase_status(string& db, vector<pair<string, PAN_PHASE_STATUS>>& phase_name_status) = 0;
	virtual string pe_get_phase_name_status(pair<string, string>& db, map<int, pair<string, PAN_PHASE_STATUS>>& phase_name_status) = 0;
	virtual string pe_get_phase_name_status(string& db, map<int, pair<string, PAN_PHASE_STATUS>>& phase_name_status) = 0;

	virtual string pe_set_phase_status(pair<string, string>& db, vector<pair<string, PAN_PHASE_STATUS>>& phase_name_status) = 0;
	virtual string pe_set_phase_status(string& db, vector<pair<string, PAN_PHASE_STATUS>>& phase_name_status) = 0;

	virtual string pe_get_system_phase_property_names(string& db, vector<string>& phase_prop_name, vector<string>& system_prop_name) = 0;

	virtual string pe_amount_converter(const string& from_unit, shared_ptr<P_Statespace> st) = 0;
	virtual string pe_amount_converter(const string& from_unit, shared_ptr<Pan_Statespace> from_st, const string& to_unit, shared_ptr<Pan_Statespace> to_st) = 0;
	// convert amount from unit of "from_unit" between x and wt

	///  add a constant value to Gibbs of a phase
	virtual string pe_add_value_to_G(string& phaseName, double deltaG_perMolarSite) = 0;

	virtual string pe_set_calculation_condition(Pan_Calculation& calc) = 0;

	virtual string pe_get_initial_point(vector<shared_ptr<P_Point>>& p_pt) = 0;

	virtual string pe_calc_point_global(shared_ptr<P_Point> p_pt, int id = 0) = 0;
	// globally stable phase equilibrium

	virtual string pe_calc_point_global_with_initial_point(shared_ptr<P_Point> p_pt) = 0;
	// globally stable phase equilibrium with initial point

	virtual string pe_calc_point_local_with_initial_point(shared_ptr<P_Point> p_pt, bool given_f = false) = 0;
	// local phase equilibrium with initial point

	virtual string pe_find_liquidus_surface(const string& liquid_phase_name, shared_ptr<P_Point> p_pt) = 0;
	// find liquidus surface

	virtual string pe_find_phase_transitions(const string& phase_name, vector<shared_ptr<P_Point>> p_pt, pair<double, double> temp_range = make_pair(2000, 300)) = 0;
	// find all phase trasitions with phase_name

	virtual string pe_calc_liquidus_slope(const string& liquid_phase_name, const string& solvent_comp_name, shared_ptr<P_Point> p_pt, const string& solid_phase_name = string("")) = 0;
	// calculate liquidus slopes in the directions of non solvent component

	virtual string pe_solidification_simulation(Solidification_Parameter& s_param, Pan_Calculation& calc, vector<Solidification_Node>& solidification_result) = 0;
	// solidifcation simulation with callback function in s_param for each step

	virtual string pe_update_properties(shared_ptr<P_Point> p_pt, bool update_phase_point_equilibrium=false, vector<string> prop_v = vector<string>()) = 0;

	virtual string pe_update_Vm_density_from_elemets(shared_ptr<P_Point> p_pt) = 0;
	// update Vm and density from elements only

	virtual string pe_calc_point_local_phase_boundary(shared_ptr<P_Point> L_p_pt, shared_ptr<P_Point> R_p_pt) = 0;

	virtual string pe_update_thermodynamic_properties(shared_ptr<P_Point> p_pt, bool update_phase_point_equilibrium=false) = 0;

	virtual string pe_set_calc_property_type(vector<pair<PAN_CALC_PROP, bool>> calc_prop) = 0;

	// debug file
	virtual string pe_open_debug(string& file_name) = 0;
	virtual string pe_close_debug() = 0;

	virtual string pe_get_database_name_stamp(vector<pair<string, string>>& db_name_stamp) = 0;
	// db_name_stamp.first = name of current database;  db_name_stamp.second = timestamp of current database

	virtual string pe_get_component_names_id(pair<string, string>& db, vector<pair<string, int>>& comp_names_id, bool active_only = false) = 0;

	virtual string pe_get_component_id_names(pair<string, string>& db, map<int, string>& comp_id_names, bool active_only = false) = 0;
	
};

// define and delete PanEngine pointer
extern "C" PANENGINE_API PanEngine* definePanEngineUser(char* msg);
extern "C" PANENGINE_API void deletePanEngineUser(PanEngine *pUser);
