
/////////////////////////////////////
//                                 //
//         PanEngine API           //
//                                 //
//         Version 2020            //
//        CompuTherm, LLC          //
//       Copyright, 2012-2020      //
//                                 //
//        www.computherm.com       //
//                                 //
/////////////////////////////////////

#pragma once

#include "Pan_Global_Def.h"


class Pan_Solidifcation_Info
{
	enum Dendrite_Shape {S_SPHARE, S_CYLINDA };
public:
	Pan_Solidifcation_Info() { m_cur_d = 0; theta = 0.5; time = 0; time_starts_at_liquidus = true; time_liq_surface = 10 + 10; }
	virtual ~Pan_Solidifcation_Info(){;}

public:
	vector<int> m_liq_ID;
	double m_dd;	// dentrite distance
	Dendrite_Shape m_shape;	// dendrite shape

public:
	// back diffusion

	map <double, pair<double, double>> m_thermal_history;
	//    t (sec)      T(K)    dT/dt

	double dTdt;
	double dT;
	double dt;
	double time;
	bool time_starts_at_liquidus;
	double time_liq_surface;	// time when liquid starts solidifying
	double theta;	// 0.5 for CN method

	map<string, vector<double>> x_prime_ph;	// concentrations of solutes in primary phase
	//  comp          a, b, c, d				// vector = <a, b, c>  Cs = a + b*x + c*x^2 + d*x^3

	string m_primary_phase;
	string m_secondary_phase;
	double m_secondary_dendrite_arm_spacing_0;
	double m_secondary_dendrite_arm_spacing_cur;
	double m_coarsening_constant;

	// solidification simulation parameters
	double m_solidification_rate;
	double m_T_gradient;
	double m_interfacial_energy;
	double m_latent_heat;
	double m_coarsening_geometric_factor;
	double m_solute_trapping_parameter;
	double m_sound_velocity;
	double m_dendrite_tip_factor;
	double m_solid_diffusivity_factor;
	double m_boundary_layer_factor;

	vector<string> m_comp;	// comp[0] is the solvent component

	map<string, double> m_primary_comp_x_average;	// average compositions in primary phase
	map<string, double> m_primary_comp_x_interface;	// interface compositions in primary phase
	map<string, double> m_primary_comp_x_first_node;	// compositions in primary phase at the first node
	map<string, double> m_primary_dxdz_interface;	// gradient of composition at interface in primary phase
	map<string, double> m_primary_J_interface;	// diffusion flux of composition at interface in primary phase
//	map<string, map<string, double>> m_primary_D_interface;	// diffusivity at interface in primary phase

	double m_cur_d;	// half of secondary dendrite arm spacing length, ds

public:
	void print_debug(ofstream& out) {
		out << "-----------------------------------\n";
		out << "**  Solidification_Info  **\n";

		out << "components: ";
		for (auto i=m_comp.begin(); i!=m_comp.end(); ++i) {
			out << (*i) << "  ";
		}
		if (m_comp.size()>0) {
			out << endl;
			out << "  solvent comp = [" << m_comp[0] << "]\n";
		}

		if (!m_primary_phase.empty()) {
			out << "primary_phase=" << m_primary_phase << endl;
		}

		if (!m_secondary_phase.empty()) {
			out << "secondary_phase=" << m_secondary_phase << endl;
		}

		if (m_secondary_dendrite_arm_spacing_cur!=0) {
			out << "m_secondary_dendrite_arm_spacing_0=" << m_secondary_dendrite_arm_spacing_0 << "m" << endl;
			out << "m_secondary_dendrite_arm_spacing_cur=" << m_secondary_dendrite_arm_spacing_cur << "m" << endl;
		}
		
		if (m_coarsening_constant!=0) {
		//	out << "coarsening_constant=" << m_coarsening_constant << "m/s^(1/3)" << endl;
			out << "coarsening_constant=" << m_coarsening_constant << "m^3/s" << endl;
		}

		// solidification simulation parameters
		if (m_solidification_rate!=0) {
			out << "solidification_rate=" << m_solidification_rate << " " << endl;
		}

		if (m_T_gradient!=0) {
			out << "T_gradient=" << m_T_gradient << " " << endl;
		}

		if (m_interfacial_energy!=0) {
			out << "interfacial_energy=" << m_interfacial_energy << " " << endl;
		}

		if (m_latent_heat!=0) {
			out << "latent_heat=" << m_latent_heat << " " << endl;
		}

		if (m_coarsening_geometric_factor!=0) {
			out << "coarsening_geometric_factor=" << m_coarsening_geometric_factor << " " << endl;
		}

		if (m_solute_trapping_parameter!=0) {
			out << "solute_trapping_parameter=" << m_solute_trapping_parameter << " " << endl;
		}

		if (m_sound_velocity!=0) {
			out << "sound_velocity=" << m_sound_velocity << " " << endl;
		}

		if (m_dendrite_tip_factor!=0) {
			out << "dendrite_tip_factor=" << m_dendrite_tip_factor << " " << endl;
		}

		if (m_solid_diffusivity_factor!=0) {
			out << "solid_diffusivity_factor=" << m_solid_diffusivity_factor << " " << endl;
		}

		if (m_boundary_layer_factor != 0) {
			out << "boundary_layer_factor=" << m_boundary_layer_factor << " " << endl;
		}
		// end of solidification simulation parameters


		out << "\nThermal History:" << endl;
		out << "t (sec)  T(K)   dT/dt\n";
		for (auto i=m_thermal_history.begin(); i!=m_thermal_history.end(); ++i) {
			out << (*i).first << "  " << (*i).second.first << "  " << (*i).second.second << endl;
		}

		out << "\ndTdt = " <<dTdt << " K/s   dT = " << dT << " K   dt=" << dt << " s" << endl;


		out << "\nm_primary_comp_x_interface:" << endl;
		for (auto i=m_primary_comp_x_interface.begin(); i!=m_primary_comp_x_interface.end(); ++i) {
			out << "x_interface[" << (*i).first << "] = " << (*i).second << endl;
		}

		out << "\nm_primary_comp_x_average:" << endl;
		for (auto i=m_primary_comp_x_average.begin(); i!=m_primary_comp_x_average.end(); ++i) {
			out << "x_average[" << (*i).first << "] = " << (*i).second << endl;
		}
		out << endl;

		out << "\nm_primary_J_interface:" << endl;
		for (auto i=m_primary_J_interface.begin(); i!=m_primary_J_interface.end(); ++i) {
			out << "J[" << (*i).first << "] = " << (*i).second << endl;
		}
		out << endl;

		//out << "\nm_primary_D_interface:" << endl;
		//for (auto i=m_primary_D_interface.begin(); i!=m_primary_D_interface.end(); ++i) {
		//	for (auto j=(*i).second.begin(); j!=(*i).second.end(); ++j) {
		//		out << "D[" << (*i).first << "," << (*j).first << "] = " << (*j).second << endl;

		//	}
		//	out << endl;
		//}
		//out << endl;
		
		out << "\nm_cur_d = " << m_cur_d * 1e+6 << " um" << endl;

		out << "\nx_prime_ph:" << endl;
		for (auto i=x_prime_ph.begin(); i!=x_prime_ph.end(); ++i) {
			out << "[" << (*i).first << "] : " << ((*i).second)[0] << "   " << ((*i).second)[1] << "   " << ((*i).second)[2] << endl;
			auto k = m_primary_dxdz_interface.find((*i).first);
			if (k != m_primary_dxdz_interface.end()) {
				out << "  dxdz[" << (*i).first << "]=" << (*k).second << endl;
			}
		}
		out << endl;

	}
};

class Solidification_Parameter {
public:
	// Jan 24 2019
	Solidification_Parameter() : m_liquid_name("Liquid"), m_primary_phase(""), m_solvent_name(""), instant_update(true), residule_liq(false), coordinate(0),
		useAverageCompositionInSolidForScheil(true), factor(1.0) {
		progress_ptr=0;
		info_ptr = make_shared<Pan_Solidifcation_Info>();
	};

	~Solidification_Parameter() {};

public:
	string m_liquid_name;
	string m_primary_phase;
	string m_solvent_name;
	int step;
		
	int liq_ID;
	set <int> liqID;
	int model_type;	// 0 for Scheil,  1 for Lever rule, 2 for BD, 100 for user_model
	int coordinate;	// 0 for cartesian, 1 for cylindrical, 2 for spherical

	double pre_lamda, cur_lamda;
	double init_tot_moles, cur_tot_moles;

	double min_liq_amount;
	double min_liq_lamda;
	double min_x;
	double T_start, T_end;	// user-defined T range
	double dT_min, dT_max;	// T step limits

	double dt_min, dt_max;	// t step limits
	double pre_t, cur_t, dt;

	double pre_T, cur_T, dT;
	double T_liq_surface;

	double T_range;	// T_range=T_start-T_end
	bool instant_update;

	bool found_transformation;
	bool residule_liq;
	
	vector <shared_ptr<P_Point>> transf_pt;
	shared_ptr<P_Point> phase_info_pt;
	bool startFromLiquidusSurface;
	bool endWhenNoMoreLiquid;
	bool useAverageCompositionInSolidForScheil;

	Pan_Progress progress_ptr;
	shared_ptr<Pan_Solidifcation_Info> info_ptr;

	//// back diffusion
	//map <double, pair<double, double>> m_cooling_curve;
	////    t (sec)      T(K)    dT/dt

	//double dTdt;

	map<string, pair<double, double>> comp_amount_0;	// <name, <x, moles>> 
//	double m_cur_liq_length;
//	double m_cur_solid_length;
	double factor;

public:
	void reset() {
		liq_ID = -1;
		startFromLiquidusSurface = true;
		endWhenNoMoreLiquid = false;
		useAverageCompositionInSolidForScheil = true;

		m_primary_phase = "";
		m_solvent_name = "";

		pre_lamda = cur_lamda = 1;
		step = 0;
		//liqID = -1;
		liqID.clear();
		model_type = 0;
		T_start = 2000;
		T_end = 300;
		dT_max = 10;
		dT_min = 0.1;
		dT = dT_min;

		dt_max = 40;
		dt_min = 0.1;
		dt = dt_min;

		pre_T = cur_T = T_start;
		dT = dT_min;
		min_liq_amount = 1e-4;
		min_liq_lamda = 0.8;
		min_x = 1e-6;
		T_liq_surface = T_start;
		instant_update = true;
		residule_liq = false;
		useAverageCompositionInSolidForScheil = true;
		transf_pt.clear();
	}

	void print_debug(ofstream& out) {
		out << "-----------------------------------\n";
		out << "**  Solidification_Parameter  **\n";
		out << "coordinate=" << coordinate;
		if (coordinate == 0) {
			out << "  (plate)";
		}
		else if (coordinate == 1) {
			out << "  (cylinder)";
		}
		else if (coordinate == 2) {
			out << "  (sphere)";
		}
		else {
			out << "  (Not defined)";
		}
		out << endl;

		out << "model_type=" << model_type;
		if (model_type == 0) {
			out << "  (Scheil)";
		} else
		if (model_type == 1) {
			out << "  (Lever rule)";
		}
		else
		if (model_type == 2) {
			out << "  (BD - back diffusion)";
		} else 
		if (model_type == 100) {
			out << "  (user_model)";
		} else {
			out << "  (Not defined)";
		}
		out << endl;

		out << "solvent=" << m_solvent_name << endl;

		out << "Step=" << step << endl;
		out << " liqID=";
		for(auto it=liqID.begin(); it!= liqID.end(); it++)
		{
		    out << "[" << *it << "] ";
        }
		out << endl;
		out << "min_liq_amount=" << min_liq_amount << endl;
		out << "min_liq_lamda=" << min_liq_lamda << endl;
		out << "min_x=" << min_x << endl;
		out << "pre_lamda=" << pre_lamda << "  cur_lamda=" << cur_lamda << endl;
		out << "T_start=" << T_start << "  T_end=" << T_end << endl;
		out << "dT_min=" << dT_min << "  dT_max=" << dT_max << endl;
		out << "dt_min=" << dt_min << "  dt_max=" << dt_max << endl;
		out << "pre_T=" << pre_T << "  cur_T=" << cur_T << "  dT=" << dT << endl;
		out << "instant_update=" << boolalpha << instant_update << endl;
		out << "residule_liq=" << boolalpha << residule_liq << endl;
		info_ptr->print_debug(out);
		out << "-----------------------------------\n";
	}

};

//class Phase_Info {
//public:
//	Phase_Info() {};
//	~Phase_Info() {};
//	Phase_Info(const Phase_Info &info) {
//		ftot = info.ftot;
//		fwtot = info.fwtot;
//		comp_tot = info.comp_tot;
//		residual_liq = false;
//	}
//
//public:
//	double ftot;	// cumulative molar fraction of the phase
//	double fwtot;	// cumulative weight fraction of the phase
//	bool residual_liq;
//	map<string, pair<double, double> > comp_tot;	// cumulatively average molar fraction (first) and weight fraction (second) of component in the phase
//	//   comp_name     x       wt
//public:
//	void print_debug(ofstream& out) {
//		out << "  ftot = " << ftot << "  fwtot=" << fwtot << endl;
//		for (auto c=comp_tot.begin(); c!=comp_tot.end(); ++c) {
//			out << "  x[" << (*c).first << "]=" <<  (*c).second.first << "  w[" << (*c).first << "]=" <<  (*c).second.second << endl;
//		}
//	}
//};

class Solidification_Node {
public :
	Solidification_Node() {
		solidLength = 0;
		lamda = 1.0;
		sdas = 0;
		fwl = 1;
		liqID.clear();
		dfs = fws = G = H = S = Cp = Vm_solid = Vm_liquid = Vm_average = H_Latent = 0;
		mole_solid = mole_liq = 0;
	}

	~Solidification_Node() {};

	Solidification_Node(const Solidification_Node &node) {
		lamda = node.lamda;
		liqID = node.liqID;
		liq_ID = node.liq_ID;

		dfs = node.dfs;
	//	fs = node.fs;
		fws = node.fws;
	//	fl = node.fl;
		fwl = node.fwl;
		G = node.G;
		H = node.H;
		S = node.S;
		Cp = node.Cp;
		Vm_solid = node.Vm_solid;
		Vm_liquid = node.Vm_liquid;
		Vm_average = node.Vm_average;
		sdas = node.sdas;
		H_H0 = node.H_H0;
		solidLength = node.solidLength;
		H_Latent = node.H_Latent;
		mole_solid = node.mole_solid;
		mole_liq = node.mole_liq;

		pt = node.pt;
		phase_info = node.phase_info;
		if (node.pt_tot!=NULL) {
			pt_tot = make_shared<P_Point>(node.pt_tot);
		}
	}

	void store(string sModel, P_Point* ppt) {
		pt = *ppt;
/*		if (s_odel == SCHEIL) {
			lamda_old = lamda;
			out << "\n&?B lamda = " << lamda << "   pt->phasePt[0].fraction = " << pt->phasePt[0].fraction << endl;
			lamda *= pt->phasePt[0].fraction;
			result.dfs = lamda_old - lamda;
		} else {
			lamda = pt->phasePt[0].fraction;
			result.dfs = 0;
		}

		result.fs = 1 - lamda;
		result.fl = lamda;
		result.Hm = pt->phasePt[liqID].H;
		pt->fs = result.fs;
		pt->fl = result.fl;
		*/
	}

public :
	double lamda;	// fraction of liquid phase remained relative to the whole system
	int liq_ID;
	set<int> liqID;	// the indices of the phasePoints which are liquid phases
	double dfs, G, H, S, Cp, Vm_solid, Vm_liquid, Vm_average;
	double fwl, fws;	// for weight fraction
	double sdas;	// second dendrite arm spacing
	double H_H0;
	double solidLength;
	double mole_solid;
	double mole_liq;

	double H_Latent; // latent heat
	// dfs : solid molar fraction solidified in this step
	// fs  : solid molar fraction solidified so far (including this step)
	// fl  : liquid molar fraction remained (after this step)
	// G  : summation of G for the system (solid phases + liquid phase)
	// H  : summation of H for the system (solid phases + liquid phase)
	// S  : summation of S for the system (solid phases + liquid phase)
	// Cp  : summation of Cp for the system (solid phases + liquid phase)
	// Vm  : summation of Vm for the system (solid phases + liquid phase)
	// H_Latent : latent heat

	P_Point pt;
	map<string, Phase_Info, icomp> phase_info;	// Jan 24 2019
	// phase_name  info

	map<string, pair<double, double> > comp_amount;	// total amounts in mole and weight for components in system
	//  comp_name  x_total  wt_total

	shared_ptr<P_Point> pt_tot;

public:
	string verify_phase_info() {
		if (phase_info.empty()) return "phase_info is empty";

		double ftot=0;
		double fwtot=0;

		// verify total
		for (auto pi=phase_info.begin(); pi!=phase_info.end(); ++pi) {
			ftot+=(*pi).second.ftot;
			fwtot+=(*pi).second.fwtot;
		}

		stringstream ostr;
		if (fabs(ftot-1) > 1e-5) {
			ostr << "ftot!=1  ftot=" << ftot << "  ";
//			return "ftot!=1  ftot="+ostr.str();
		}

		if (fabs(fwtot-1) > 1e-5) {
	//		stringstream ostr;
			ostr << "fwtot!=1  fwtot=" << fwtot << "  ";
	//		return "fwtot!=1  fwtot="+ostr.str();
		}


		if (fabs(ftot-1) > 1e-5 || fabs(fwtot-1) > 1e-5) return "Error: "+ostr.str();

		// verify components
		set <string> comp;
		for (auto pi=phase_info.begin(); pi!=phase_info.end(); ++pi) {
			for (auto c=(*pi).second.comp_tot.begin(); c!=(*pi).second.comp_tot.end(); ++c) {
				comp.insert((*c).first);
			}
		}

		double sum_x=0;
		double sum_w=0;
		for (auto c_name=comp.begin(); c_name!=comp.end(); ++c_name) {
			for (auto pi=phase_info.begin(); pi!=phase_info.end(); ++pi) {
				//	map<string, pair<double, double> > comp_tot;
				auto c=(*pi).second.comp_tot.find(*c_name);
				if (c!=(*pi).second.comp_tot.end()) {
					sum_x+=(*c).second.first * (*pi).second.ftot;
//					cout << "sum_x=" <<sum_x << "  (*c).second.first=" << (*c).second.first << "    (*pi).second.ftot=" << (*pi).second.ftot << endl;
					sum_w+=(*c).second.second * (*pi).second.fwtot;
				}
			}
		}

		if (fabs(sum_x-1) > 1e-5) {
			stringstream ostr;
			ostr << sum_x;
			return "sum_x-1!=0  sum_x="+ostr.str();
		}

		if (fabs(sum_w-1) > 1e-5) {
			stringstream ostr;
			ostr << sum_w;
			return "sum_w-1!=0  sum_w="+ostr.str();
		}
		return "";
	}

	string update_totoal_amount_of_component() {
		if (phase_info.empty()) return "phase_info is empty";

		comp_amount.clear();

		// verify components
		set <string> comp;
		for (auto pi=phase_info.begin(); pi!=phase_info.end(); ++pi) {
			for (auto c=(*pi).second.comp_tot.begin(); c!=(*pi).second.comp_tot.end(); ++c) {
				comp.insert((*c).first);
			}
		}

		for (auto c_name=comp.begin(); c_name!=comp.end(); ++c_name) {
			double sum_x=0;
			double sum_w=0;
			for (auto pi=phase_info.begin(); pi!=phase_info.end(); ++pi) {
				//	map<string, pair<double, double> > comp_tot;
				auto c=(*pi).second.comp_tot.find(*c_name);
				if (c!=(*pi).second.comp_tot.end()) {
					sum_x+=(*c).second.first * (*pi).second.ftot;
//					cout << "sum_x=" <<sum_x << "  (*c).second.first=" << (*c).second.first << "    (*pi).second.ftot=" << (*pi).second.ftot << endl;
					sum_w+=(*c).second.second * (*pi).second.fwtot;
				}
			}
			comp_amount.insert(pair<string, pair<double, double>>(*c_name, make_pair(sum_x, sum_w)));
		}

		return "";
	}

	void print_debug(ostream& out) {
		out << "mole_solid = " << mole_solid << "  mole_liq=" << mole_liq << "  Vm_solid = " << Vm_solid << "  Vm_liquid = " << Vm_liquid << "  Vm_average = " << Vm_average << endl;
		out << "pt:";
		pt.print_debug(out);

		out << "H_H0=" << H_H0 << "  H_Latent=" << H_Latent << endl;
		out << "[Phase Info]\n";
		if (phase_info.empty()) {
			out << "<Empty>\n";
		} else {
			for (auto ph=phase_info.begin(); ph!=phase_info.end(); ++ph) {
				if ((*ph).second.residual_liq)
				{
					out << (*ph).first << " {residual}" << ":\n";
				}
				else
				{
					out << (*ph).first << ":\n";;
				}
				(*ph).second.print_debug(out);
			}

			string ok=verify_phase_info();
			update_totoal_amount_of_component();

			out << "Total Amount:" << endl;
			for (auto c=comp_amount.begin(); c!=comp_amount.end(); ++c) {
				out << "  " << (*c).first << "  x=" << (*c).second.first << "  w=" << (*c).second.second << endl;
			}
			out << "  Vm = " << Vm_average << endl;

			//if (ok.empty())	out << "phase info is consistent\n";
			//else out << "phase info is NOT consistent : " << ok << "\n";
			if (!ok.empty()) out << "phase info is NOT consistent : " << ok << "\n";

			if (pt_tot != NULL) {
				out << "\npt_tot (Point with average composition):" << endl;
				pt_tot->print_debug(out);
			}
		}
	}

	void print_debug_short(ostream& out) {
		out << "pt:\n";
		pt.print_debug_short(out);
	}

};


